//
//  wbiff! mailbox flag for Windows
//	Copyright  1999, 2000 F. Sadaghiani (sada@guesswork.org)
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 
//
//  59 Temple Place - Suite 330
//  Boston, MA 
//  02111-1307, USA.
//
//

#include "std_wb.h"
#include "types.h"

#include "impl.thread.h"

#include "impl.h"
#include "mail.general.h"
#include "mail.pop3.h"
#include "mail.imap4.h"
#include "skin.h"
#include "wb_util.h"
#include "res/resource.h"

#define IDI_ANIM_1     IDI_ICON2
#define IDI_ANIM_2     IDI_ICON3
#define IDI_ANIM_3     IDI_ICON4
#define IDI_ANIM_4     IDI_ICON5
#define IDI_BLANK      IDI_ICON6

extern HINSTANCE hInstance;
extern SKIN skin;
extern SETTINGS settings;
extern STATS stats;
extern HANDLE hThread_CheckNow;
extern HANDLE hThread_FlashWindow;
extern HANDLE hThread_FlashTray;
extern HANDLE hThread_CheckMailAnimation;
extern HANDLE hEvent_CheckMailAnimation;
extern HANDLE hEvent_ContinueAnimationTray;
extern HANDLE hEvent_ContinueAnimationWindow;
extern int nCount;

extern DWORD dwState;

static void
CallGetCountProc(int nIndex)
{
	//
	// Calls correct GetCount function based on account's protocol
	// specification
	//

	typedef void (*MGCF)(MAIL_PACCOUNT, bool);

	MGCF mail_GetCount = NULL;

	switch (settings.accounts[nIndex].protocol) {

		case p_POP3:
			mail_GetCount = mail_pop3_GetCount;
			break;

		case p_IMAP4:
			mail_GetCount = mail_imap4_GetCount;
			break;
	}

	if (mail_GetCount)
		mail_GetCount(&settings.accounts[nIndex], settings.bObtainEmailHeaders);
}

DWORD WINAPI
impl_ThreadProc_GetCount(LPVOID lpParameter)
{
	CallGetCountProc((int) lpParameter);
	return 0;
}

DWORD WINAPI 
impl_ThreadProc_CheckNow(LPVOID lpParameter)
{
	// Note: When checking mail the user cannot modify the program
	//       settings so it's safe to check (and modify) the settings
	//       variable here. (i.e. no need for a critical section)
	//
	// depending on the type of check occurring (i.e. linear or parallel)
	// different numbers of threads are spun. This thread function will
	// spin one thread for each account to check mail on. If the user has
	// chosen to check all accounts in parallel, for each account, a new
	// thread is spun simultaneously. Conversely, if the check is to occur
	// linearly, one thread is spun and terminates before the next one is
	// spun.
	//
	// CHANGE: Linear checks no longer spin an extra thread to do the check.
	//         All checks are done in the current thread.

	const HWND hwnd = (HWND) lpParameter;

	DWORD tid;
	int i, count = 0;
	bool success = false; // was there at least one success?

	if (settings.bCheckInParallel && (settings.nAccounts > 1)) {

		HANDLE* threads = wb_malloc(settings.nAccounts * sizeof(HANDLE));
		if (threads) {

			for (i = 0; i < settings.nAccounts; i++)
				threads[i] = wb_thread(impl_ThreadProc_GetCount, i, &tid);

			WaitForMultipleObjects(settings.nAccounts, threads, TRUE, INFINITE);

			for (i = 0; i < settings.nAccounts; i++) {

				CloseHandle(threads[i]);

				if (settings.accounts[i].nCount >= 0) {

					count += settings.accounts[i].nCount;
					success = true;
				}
			}

			wb_free(threads);
		}
	}
	else { // perform e-mail check linearly

		for (i = 0; i < settings.nAccounts; i++) {

			CallGetCountProc(i);

			// update count and success
			if (settings.accounts[i].nCount >= 0) {

				count += settings.accounts[i].nCount;
				success = true;
			}
		}
	}

	++stats.nTimesChecked;

	PostMessage(hwnd, WBIFF_IDM_THREAD_DONE_CHECK_NOW, (WPARAM) success, count);
	return 0;
}

void 
OnThreadDone_CheckNow(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	TCHAR tbuffer[64]; // aligned
	
	int nOldCount = nCount;
	bool bFlagWasDown = !(dwState & WBIFF_CURRENT_STATE_FLAG_UP);
	
	if (settings.bCheckMailAnimation) {

		SetEvent(hEvent_CheckMailAnimation);
		WaitForSingleObject(hThread_CheckMailAnimation, INFINITE);
	}
	
	CloseHandle(hThread_CheckNow);
	hThread_CheckNow = NULL;


	impl_ResetFlagAndTrayState(hwnd);
	

	// get time stamp for window/tray text
	GetTimeFormat(LOCALE_USER_DEFAULT, LOCALE_NOUSEROVERRIDE | TIME_NOSECONDS, NULL,
		NULL, tbuffer, sizeof(tbuffer));

	if (!wParam) {

		// there were no successful checks made; attempt to obtain e-mail
		// count failed (for every server)

		impl_UpdateWindowText(_T("Unable to obtain e-mail counts"));
		impl_UpdateToolTipText(_T("Unable to obtain e-mail counts"));
		impl_UpdateSummaryWindow(hwnd, false);
		InvalidateRect(hwnd, &skin.FontRect, TRUE);
		UpdateWindow(hwnd);
	}
	else if (lParam == 0) {

		// no new messages (may have some errors)
		impl_UpdateWindowText(_T("No new messages; checked at %s"), tbuffer);
		impl_UpdateToolTipText(_T("No new messages; checked at %s"), tbuffer);
		impl_UpdateSummaryWindow(hwnd, false);
		InvalidateRect(hwnd, &skin.FontRect, TRUE);
		UpdateWindow(hwnd);
	}	
	else if (lParam == nOldCount) {

		// No change in the number of e-mails waiting for the user
		// (Even though the number of messages hasn't changed, we must
		//  update the last "checked @ ..." time to indicate to the user
		//  that the program is in fact working)
		
		impl_UpdateWindowText(_T("%d waiting;\nchecked at %s"), nOldCount, tbuffer);
		impl_UpdateToolTipText(_T("%d waiting; checked at %s"), nOldCount, tbuffer);

		// if the user chose to "Clear Flag" then they don't wish to be notified
		// when the count is the same as during the previous check(s)
		if (bFlagWasDown) {

			impl_UpdateSummaryWindow(hwnd, false);
			InvalidateRect(hwnd, &skin.FontRect, TRUE);
			UpdateWindow(hwnd);
		}
		else {

			// set the flag and tray signals in the up state
			impl_SetFlagAndTrayState(hwnd);

			// if user has chosen to display summary window (HEADERS tab) then
			// display that too.
			impl_UpdateSummaryWindow(hwnd, true);
		}
	}
	else {

		// update program stats
		if (lParam > nCount) {
			
			// some new messages have arrived (some already existed in nCounts count of
			// messages)
			stats.nMessages += lParam - nCount;
		}
		else { // lParam < nCount

			// assume that the new count lParam is less than nCount because
			// the user checked his inbox w/o the program knowing about it and
			// the new count is the total count of messages waiting in the inbox
			stats.nMessages += lParam;
		}

		// New messages (could be less or more) have arrived in mailbox(es)
		impl_UpdateWindowText(_T("%d waiting;\nchecked at %s"), lParam, tbuffer);
		impl_UpdateToolTipText(_T("%d waiting; checked at %s"), lParam, tbuffer);
		impl_UpdateSummaryWindow(hwnd, true);
		impl_SetFlagAndTrayState(hwnd);

		// if we need to notify the user, do so now.
		if (lParam >= settings.nMinLimit) {

			// play the sound file
			if (settings.bPlayWave && settings.szWaveFile[0] != 0)
				PlaySound(settings.szWaveFile, NULL, SND_NODEFAULT | SND_NOSTOP | SND_ASYNC | SND_FILENAME | SND_NOWAIT);

			if (settings.bPopupWindow && !(dwState & WBIFF_CURRENT_STATE_MAINWND_VISIBLE)) {
				
				impl_ToggleMainWindow(hwnd);
			}

			if ((settings.bFlashWindow && (dwState & WBIFF_CURRENT_STATE_MAINWND_VISIBLE)) || settings.bFlashTray)
				PostMessage(hwnd, WBIFF_IDM_BEGIN_NOTIFY_ANIMATION, 0, 0);

			// Run client
			if (settings.bRunClient) {

				// We don't call `impl_RunClient' because we don't want to
				// reset the flag and tray state
				util_Exec(hwnd, settings.szEmailClient, settings.szEmailClientArgs);
			}
		}
	}

	nCount = lParam; // update count with new count

// ----------------------------------------------------------------------

	if (!settings.bIgnoreErrors) {

		int i;
		TCHAR szErrorText[1024];

		// after the visual indicators, we display error message boxes
		// for each account that caused a problem

		for (i = 0; i < settings.nAccounts; ++i) {

			if (settings.accounts[i].nCount < 0) {

				// display the error this account had
				wsprintf(szErrorText, 
					_T("An error occurred while attempting to obtain the e-mail count on \"%s (%s, port %d)\".\n\nError: (%d) %s\nServer response: %s"),
					settings.accounts[i].szDisplayName,
					settings.accounts[i].szHost,
					settings.accounts[i].nPort,
					settings.accounts[i].nCount,
					mail_general_GetErrorMessageFromCode(settings.accounts[i].nCount),
					settings.accounts[i].szErrorText);

				MessageBox(hwnd, szErrorText, WBIFF_WINDOWNAME, MB_OK | MB_ICONSTOP);
			}
		}
	}

	// Remove the checking mail state
	dwState &= ~WBIFF_CURRENT_STATE_CHECKING_MAIL;

	// Only re-enable the timer if the user is in "Enabled" mode
	if (dwState & WBIFF_CURRENT_STATE_ENABLED)
		impl_SetTimer(hwnd, true);

	impl_SwapOutOfMemory();
}

DWORD WINAPI 
impl_ThreadProc_FlashTray(LPVOID lpParameter)
{
	const HWND hwnd = (HWND) lpParameter;

	bool bCleanUp = false;
	HICON hIconUp = skin.hIconFlagUp;
	HICON hIconDown = skin.hIconFlagDown;

	if (settings.bFlashFlagDownOnly) {

    // load the "blank" icon for animation
		hIconDown = (HICON) LoadImage(hInstance, MAKEINTRESOURCE(IDI_BLANK), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
		bCleanUp = true;
	}
	
	// flashing tray animation
	if (hIconUp && hIconDown) {

		NOTIFYICONDATA tray;
		int i = 0;
		bool bDrawDown = true;

		tray.cbSize = sizeof(NOTIFYICONDATA);
		tray.hWnd = hwnd;
		tray.uFlags = NIF_ICON;
		tray.uID = IDC_WBIFF_TRAYICON;
	
		while (

			(WaitForSingleObject(hEvent_ContinueAnimationTray, 0) != WAIT_OBJECT_0)

			&&
			
			(settings.bContinueAnimationTray ? true : (i++ < settings.nAnimationIterations))) {
		
			tray.hIcon = bDrawDown ? hIconDown: hIconUp;
			Shell_NotifyIcon(NIM_MODIFY, &tray);
			Sleep(settings.nAnimationDelay);

			bDrawDown = !bDrawDown;
		}

		// once the animation is complete, set the icon to the 
		// up state. (since new mail has arrived)

		tray.hIcon = hIconUp;
		Shell_NotifyIcon(NIM_MODIFY, &tray);

		TRACE("impl_ThreadProc_FlashTray: Set flag state up\r\n");
	}
	
	if (bCleanUp)
		DestroyIcon(hIconDown);

	PostMessage(hwnd, WBIFF_IDM_THREAD_DONE_FLASH_TRAY, 0, 0);
	return 0;
}

void
OnThreadDone_FlashTray(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	// all the work is done in OnThreadDone_FlashWindow handler
	CloseHandle(hThread_FlashTray);
	hThread_FlashTray = NULL;

	CloseHandle(hEvent_ContinueAnimationTray);
	hEvent_ContinueAnimationTray = NULL;


#if 0
/*
 * moved to thread handler funciton
 *
 *

	if (settings.bContinueAnimationTray) {

		NOTIFYICONDATA tray;
	
		// Reset the tray state based on the count of waiting messages
		tray.uFlags = NIF_ICON;
		tray.hWnd = hwnd;
		tray.cbSize = sizeof(NOTIFYICONDATA);
		tray.uID = IDC_WBIFF_TRAYICON;
		tray.hIcon = (!(dwState & WBIFF_CURRENT_STATE_FLAG_UP) && nCount) ? skin.hIconFlagDown : skin.hIconFlagUp;
		Shell_NotifyIcon(NIM_MODIFY, &tray);
	}
*/
#endif
}

DWORD WINAPI 
impl_ThreadProc_FlashWindow(LPVOID lpParameter)
{
	// Flash between "Flag Down" and "Flag Up" images in the
	// bitmap. Always end on the "Flag Up" image. After the
	// thread is done, invalidate the region specified by
	// skin.FlagRect.

	const HWND hwnd = (HWND) lpParameter;

	HDC hdc = GetDC(hwnd), cdc;
	int i = 0;
	bool bDrawDown = false;

	cdc = CreateCompatibleDC(hdc);
	SelectObject(cdc, skin.hBitmap);
	
	while (
			(WaitForSingleObject(hEvent_ContinueAnimationWindow, 0) != WAIT_OBJECT_0)	

			&&

			(settings.bContinueAnimationWindow ? true : (i++ < settings.nAnimationIterations))) {
	
		BitBlt(hdc,
			skin.FlagRect.left,
			skin.FlagRect.top,
			skin.FlagRect.right - skin.FlagRect.left,
			skin.FlagRect.bottom - skin.FlagRect.top,
			cdc,
			bDrawDown ? skin.FlagDown_X : skin.FlagUp_X,
			bDrawDown ? skin.FlagDown_Y : skin.FlagUp_Y,
			SRCCOPY);

		Sleep(settings.nAnimationDelay);
		bDrawDown = !bDrawDown;
	}

	// draw flag in up position
	DeleteDC(cdc);
	ReleaseDC(hwnd, hdc);

	PostMessage(hwnd, WBIFF_IDM_THREAD_DONE_FLASH_WINDOW, 0, 0);
	return 0;
}

void 
OnThreadDone_FlashWindow(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	CloseHandle(hThread_FlashWindow);
	hThread_FlashWindow = NULL;

	CloseHandle(hEvent_ContinueAnimationWindow);
	hEvent_ContinueAnimationWindow = NULL;

	// force window to repaint itself. Note that the flag and tray
	// state have been set (via a call to impl_SetFlagAndTrayState)
	// in the OnThreadDone_CheckNow *before* calling this animation
	// loop
	
	InvalidateRect(hwnd, NULL, TRUE);
	UpdateWindow(hwnd);

#if 0
	if (settings.bFlashTray && !settings.bContinueAnimationTray) {

		// wait for the tray animation to complete before
		// proceeding
		WaitForSingleObject(hThread_FlashTray, INFINITE);
	}

	// reset the flag and tray icon state if the flag has been
	// signalled down or the user has specified to launch an
	// e-mail client
	if (!(dwState & WBIFF_CURRENT_STATE_FLAG_UP) || settings.bRunClient) {

		ASSERT(nCount == 0);
		impl_ResetFlagAndTrayState(hwnd);
	}
	else {

		impl_SetFlagAndTrayState(hwnd);
		InvalidateRect(hwnd, NULL, TRUE);
		UpdateWindow(hwnd);
	}
#endif
}

DWORD WINAPI 
impl_ThreadProc_CheckMailAnimation(LPVOID lpParameter)
{
	const HWND hwnd = (HWND) lpParameter;

	HICON hIconFrames[4];
	int nIndex = 0;
	NOTIFYICONDATA tray;
	int nDelay = settings.nAnimationDelay;
	
	if (settings.bHurryCheckMailAnimation)
		nDelay /= 4;

	tray.cbSize = sizeof(NOTIFYICONDATA);
	tray.hWnd = hwnd;
	tray.uFlags = NIF_ICON;
	tray.uID = IDC_WBIFF_TRAYICON;

	hIconFrames[0] = (HICON) LoadImage(hInstance, MAKEINTRESOURCE(IDI_ANIM_1), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
	hIconFrames[1] = (HICON) LoadImage(hInstance, MAKEINTRESOURCE(IDI_ANIM_2), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
	hIconFrames[2] = (HICON) LoadImage(hInstance, MAKEINTRESOURCE(IDI_ANIM_3), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);
	hIconFrames[3] = (HICON) LoadImage(hInstance, MAKEINTRESOURCE(IDI_ANIM_4), IMAGE_ICON, 16, 16, LR_DEFAULTCOLOR);

	while (WaitForSingleObject(hEvent_CheckMailAnimation, 0) != WAIT_OBJECT_0) {

		tray.hIcon = hIconFrames[nIndex];
		if (++nIndex > 3)
			nIndex = 0;

		Shell_NotifyIcon(NIM_MODIFY, &tray);
		Sleep(nDelay);
	}

	DestroyIcon(hIconFrames[3]);
	DestroyIcon(hIconFrames[2]);
	DestroyIcon(hIconFrames[1]);
	DestroyIcon(hIconFrames[0]);

	PostMessage(hwnd, WBIFF_IDM_THREAD_DONE_CHECK_MAIL_ANIMATION, 0, 0);
	return 0;
}

void
OnThreadDone_CheckMailAnimation(HWND hwnd, WPARAM wParam, LPARAM lParam)
{
	CloseHandle(hEvent_CheckMailAnimation);
	CloseHandle(hThread_CheckMailAnimation);

	hEvent_CheckMailAnimation =
	hThread_CheckMailAnimation = NULL;
}

