//
//  wbiff! mailbox flag for Windows
//	Copyright  1999, 2000 F. Sadaghiani (sada@guesswork.org)
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 
//
//  59 Temple Place - Suite 330
//  Boston, MA 
//  02111-1307, USA.
//
//

#include "std_wb.h"
#include "types.h"

#include "mail.general.h"

#pragma intrinsic(memcpy,memset)
#pragma hdrstop

void
mail_general_FreeAccount(MAIL_PACCOUNT pAccount)
{
	if (pAccount->messages) {
		
		// free message data
		wb_free(pAccount->messages);
	}
}

// allocates space for message header information
bool
mail_general_AllocMessages(MAIL_PACCOUNT pAccount)
{
	ASSERT(pAccount->nCount > 0);
	
	mail_general_FreeAccount(pAccount);
	
	pAccount->messages = (MAIL_PMESSAGE) wb_malloc(pAccount->nCount * sizeof(MAIL_MESSAGE));
	if (pAccount->messages) {

#if 1
		// clear message structure
		int i;
		for (i = 0; i < pAccount->nCount; ++i) {
			
			ZeroMemory(pAccount->messages[i].szDate, MAIL_ACCOUNT_FIELD_LENGTH);
			ZeroMemory(pAccount->messages[i].szFrom, MAIL_ACCOUNT_FIELD_LENGTH);
			ZeroMemory(pAccount->messages[i].szSubject, MAIL_ACCOUNT_FIELD_LENGTH);
		}
#endif

		return true;
	}
	else
		return false;
}

void
mail_general_FreeAllAccounts(MAIL_PACCOUNT pAccounts, int nCount)
{
	int i;

	for (i = 0; i < nCount; ++i)
		mail_general_FreeAccount(&pAccounts[i]);
}

/*
 * Function: mail_general_GetHostAddress
 * 
 * Description
 *  This function will get host address and put it into a format such
 * that the mail_general_ConnectTo function will work. We must be able to
 * translate both text server names and IP names.
 *
 * Parameters
 *  lpHostName -> name of host server (either IP address or host lpHostName)
 *                (NOTE: lpHostName is changed in place)
 *  pAddr -> in_addr representation of host
 *
 * Return
 *  0 on success, -1 otherwise
 */
int
mail_general_GetHostAddress(LPTSTR lpHostName, struct in_addr* addr)
{
  if (*lpHostName == _T('\0')) {

    return -1;
  }
	else {

		if ((*lpHostName >= _T('0')) && (*lpHostName <= _T('9'))) {

			/* IP address (e.g. "24.112.3.21") */

#ifdef _UNICODE

			struct sockaddr_in sa_temp;
			DWORD dwSize = sizeof(SOCKADDR_IN);
			
			if (WSAStringToAddress(lpHostName, AF_INET, NULL, (LPSOCKADDR) &sa_temp, &dwSize) == SOCKET_ERROR)
				return -1;

			addr->s_addr = sa_temp.sin_addr.s_addr;

#else
			
			addr->s_addr = inet_addr(lpHostName);
			if ((long) addr->s_addr == -1) // note: addr->s_addr is unsigned
				return -1;

#endif
		}
		else {

			/* host lpHostName (e.g. "www.microsoft.com") */

#ifdef _UNICODE

			WSAQUERYSET wsaQuery;
			LPWSAQUERYSET lpResult;
			GUID svcHostAddrByName = SVCID_INET_HOSTADDRBYNAME;
			HANDLE hQuery;
			DWORD dwBufferLength = 0;

			ZeroMemory(&wsaQuery, sizeof(WSAQUERYSET));
			wsaQuery.dwSize = sizeof(WSAQUERYSET);
			wsaQuery.dwNameSpace = NS_ALL;
			wsaQuery.lpServiceClassId = &svcHostAddrByName;
			wsaQuery.lpszServiceInstanceName = lpHostName;

			if (WSALookupServiceBegin(&wsaQuery, LUP_RETURN_BLOB | LUP_RETURN_ADDR, &hQuery) == SOCKET_ERROR)
				return -1;

			WSALookupServiceNext(hQuery, 0, &dwBufferLength, NULL);

			if (!(lpResult = (LPWSAQUERYSET) wb_malloc(dwBufferLength))) {

				WSALookupServiceEnd(hQuery);
				return -1;
			}

			if (WSALookupServiceNext(hQuery, 0, &dwBufferLength, lpResult) == SOCKET_ERROR) {
				
				WSALookupServiceEnd(hQuery);
				wb_free(lpResult);
				return -1;
			}

			memcpy(&(addr->s_addr),	&((struct sockaddr_in*) lpResult->lpcsaBuffer[0].RemoteAddr.lpSockaddr)->sin_addr,	sizeof(struct in_addr));

			wb_free(lpResult);
			WSALookupServiceEnd(hQuery);

#else
	
			struct hostent* he = gethostbyname(lpHostName);

			if (!he)
				return -1;
		
			memcpy(addr, he->h_addr, sizeof(struct in_addr));

#endif
		}
	}
	
	return 0;
}

/*
 * function: mail_general_ConnectTo(char const*, int)
 * 
 * Description
 *  Connects to the server named host on the port specified by port
 *
 * Parameters
 *  pAccount - pointer to account information
 *
 * Return
 *  the file descriptor of the attached hSocket or 
 *  INVALID_SOCKET on error
 */
int
mail_general_ConnectTo(const MAIL_PACCOUNT pAccount)
{
	IN_ADDR ia;
	
  if (mail_general_GetHostAddress(pAccount->szHost, &ia) < 0)
		return INVALID_SOCKET;
	else {

		SOCKADDR_IN sa;
		SOCKET hSocket;
		
		sa.sin_addr.s_addr = ia.s_addr;
		sa.sin_family = AF_INET;
		sa.sin_port = htons(pAccount->nPort);
		
		if ((hSocket = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
			return INVALID_SOCKET;
		
		if (pAccount->nTimeout > 0) {

			int timeout = pAccount->nTimeout * 1000; // convert seconds to milliseconds

			setsockopt(hSocket, SOL_SOCKET, SO_RCVTIMEO, (LPCTSTR) &timeout, sizeof(int));
			setsockopt(hSocket, SOL_SOCKET, SO_SNDTIMEO, (LPCTSTR) &timeout, sizeof(int));
		}
		
		if (connect(hSocket, (LPSOCKADDR) &sa, sizeof(SOCKADDR_IN)))
			return INVALID_SOCKET;
		
		return hSocket;
  }
}

bool
mail_general_ReceiveSingleLine(SOCKET hSocket, LPTSTR lpBuffer, int nLength)
{
	// as defined by RFC1939, the maximum length of a single line
	// response from A POP3 server is 512 bytes/characters.

	TCHAR ch;
	int result, pos = 0;

	while (true) {

		result = recv(hSocket, &ch, 1, 0);

		// check for recv timeout
		if (result == SOCKET_ERROR && WSAGetLastError() == WSAETIMEDOUT) {

			TRACE(_T("CONNECTION TIMED OUT: SERVER FAILED TO RESPOND; throwing WSAETIMEDOUT exception\r\n"));
			return false;
		}
		else if (--nLength >= 0 && result == 1) {

			lpBuffer[pos++] = ch;

			// only stop receiving if we've hit a CRLF pair; this 
			// occurs after we've received at least 2 bytes and
			// the current character is \n and the previous character
			// is \r
			if (ch == _T('\n') && pos > 1) {

				if (lpBuffer[pos - 1 - 1] == _T('\r'))
					break;
			}
		}
		else {

			// buffer too small or error occurred while recv'ing bytes
			// from host

			// adjust pos
			--pos;
			break;
		}
	}

	lpBuffer[pos] = _T('\0');
	TRACE(_T("S:  %s"), lpBuffer);
	return true;
}

bool
mail_general_SendSingleLine(SOCKET hSocket, LPTSTR lpBuffer, int nLength)
{
#ifdef _UNICODE

	// convert Unicode string to ANSI before
	// sending to host
	
	LPSTR lpAnsiBuffer = NULL;
	if (lpAnsiBuffer = wb_malloc((nLength + 1) * sizeof(char))) {

		WideCharToMultiByte(CP_ACP, 0, lpBuffer, nLength, lpAnsiBuffer, nLength, NULL, NULL);
		send(hSocket, lpAnsiBuffer, nLength, 0);
		wb_free(lpAnsiBuffer);	
	}
#else
	// do regular send
	send(hSocket, lpBuffer, nLength, 0);
#endif

	if (WSAGetLastError() == WSAETIMEDOUT) {

		TRACE(_T("CONNECTION TIMED OUT: CLIENT FAILED TO SEND; throwing WSAETIMEDOUT exception\r\n"));
		return false;
	}

	TRACE(_T("C:  %s"), lpBuffer);
	
	return true;
}

// startup winsock
int 
mail_general_Startup(void)
{
	WORD wVersionRequested;
	WSADATA wsaData;

	wVersionRequested = MAKEWORD(2, 2);
	
	return WSAStartup(wVersionRequested, &wsaData);
}

// cleanup winsock
void
mail_general_Cleanup(void)
{
	WSACleanup();
}

LPCTSTR
mail_general_GetErrorMessageFromCode(int error)
{
	static LPCTSTR errors[] =
	{
		// general
		_T("Error trying to logon to the mail system."),
		_T("The connection attempt timed-out."),
		_T("The connection has been dropped because the remote system failed to respond."),
		_T("Unable to establish a connection to the specified host."),

		// imap4
		_T("Unable to select the inbox folder (IMAP4)."),
		_T("Unable to execute check command (IMAP4)."),
		_T("Unable to search for new messages (IMAP4)."),

		// pop3
		_T("Unable to determine message count (POP3)."),
		_T("Unable to execute TOP command (POP3)."),
	};

	return errors[error + (-ERROR_CODES_START)];
}



